#!/usr/bin/python
# -*- coding: utf-8 -*-


# Whitelist this app
#  gsettings set com.canonical.Unity.Panel  systray-whitelist "['JavaEmbeddedFrame', 'Wine', 'scp-dbus-service', 'Update-notifier', 'python']"

# go to correct working dir
import gettext
import httplib
import locale
import os
import platform
import sys
import traceback
from time import sleep

from PyQt4 import QtGui as Q
from PyQt4 import uic
from PyQt4.QtCore import pyqtSlot, QUrl, QThread, pyqtSignal, SLOT, SIGNAL, Qt

import autorun
import version
from omnicopy_api import OmniCopyAPI, ST


def log(text):
    pass
    # print(text)


print("Platform: '" + platform.system() + "'")

try:
    from pkg_resources import resource_filename
    use_pkg_resources = True
    print("Use PGK Resources")

except:
    use_pkg_resources = False
    print("No PKG Resources")


def RES(filename):
    if hasattr(sys, 'frozen'):
        if platform.system() is 'Windows':
            return sys.prefix + "\\data\\%s" % filename
        else:
            return filename
    elif use_pkg_resources:
        return resource_filename(__name__, filename)
    else:
        return filename


current_locale, encoding = locale.getdefaultlocale()
log("\nStarting OmniCopy ...")
log("Trying localization for: %s" % current_locale)
if (current_locale == None):
    current_locale = "de"
    log("  fallback to EN")

try:
    gettext.install(version.APP_NAME)
    language = gettext.translation('omnicopy', "locale", [current_locale])
except IOError:
    log("  could not Find Language file for %s" % (current_locale))
    language = gettext.NullTranslations()

language.install()
_ = language.ugettext


def toQString(text):
    return text.encode('utf-8')


def toUnicode(text):
    return unicode(text.toUtf8(), 'utf-8')


class OmniCopy(Q.QWidget):
    newClipboard = pyqtSignal(basestring)
    trayIcon = None
    api = OmniCopyAPI()
    stat = {0: _("Connected (new data)"),
            1: _("Connected"),
            2: _("No internet"),
            3: _("Not linked"),
            4: _("Linking ...")}

    def __init__(self, app):
        self.app = app;
        Q.QWidget.__init__(self)
        self.systemClipboard = Q.QApplication.clipboard()
        self.resize(250, 150)
        self.setWindowTitle('OmniCopy')
        self.createActions()
        self.createTrayIcon()
        self.trayIcon.show()
        self.clipActions = list()
        self.about_dialog = None
        self.newVersion = ""
        self.labelNewVersion = Q.QLabel(self.newVersion)
        # Setup Update Timer
        self.updateThread = UpdateThread(self.api, self.newClipboard)
        self.updateThread.updateData.connect(self.update)
        self.updateThread.start()
        self.downloadClip = ""
        self.lastMousePos = None
        self.autoUpdate()
        self.linkDb()

    def quit(self):
        log("waiting for thread to finish ...")
        self.updateThread.setRunning(False)
        self.app.closeAllWindows()
        self.trayIcon.hide()
        while self.updateThread.isRunning():
            sleep(0.05)
        log("Quit")
        self.app.exit(0)

    def updateTrayIconMenu(self):

        list = self.api.getClipboard()

        self.trayIconMenu.setUpdatesEnabled(False)

        for act in self.clipActions:
            self.trayIconMenu.removeAction(act)
            del act

        del self.clipActions[:]

        # self.trayIconMenu.insertSeparator(self.insertPoint)
        first = True
        for clip in list:
            if None == clip: continue

            if first:  # Store first clipboard for later use
                self.downloadClip = clip
                first = False

            receiver = self.factory(clip)

            if len(clip) > 100:
                clip = clip[:100] + " ... [%d]" % (len(clip))
            newAct = Q.QAction(clip.strip().replace('&', '&&').replace('\n', ' '), self, triggered=receiver)

            # newAct.setIcon(self.icon)
            newMenuItem = self.trayIconMenu.insertAction(self.insertPoint, newAct)
            self.clipActions.append(newAct)

        # self.trayIconMenu.insertSeparator(self.insertPoint)
        self.trayIconMenu.update()
        if self.trayIconMenu.isVisible():
            if None != self.lastMousePos:
                self.trayIconMenu.activateWindow()
                self.trayIconMenu.popup(self.lastMousePos)

    def update(self, state):
        self.trayIconMenu.setUpdatesEnabled(False);
        self.status.setText(self.stat[state])
        log("Status OmnniCopy: %s" % self.stat[state])

        if state == ST.NEW_DATA:
            self.trayIcon.setIcon(self.icon_bright)
            self.updateTrayIconMenu()
            self.newClipboard.emit(self.getClipboard())
            self.copyNewEntryToClipboard()
            if self.optionNotify.isChecked():
                self.showInfoMessage(self.downloadClip, _('New clipboard'))


        elif state == ST.CONNECTED:
            self.trayIcon.setIcon(self.icon_dark)
            self.newClipboard.emit(self.getClipboard())

        elif state == ST.NO_INTERNET:
            pass

        elif state == ST.NOT_AUTHORIZED:
            pass

        elif state == ST.WAIT_AUTHORIZE:  # linking handled by thread
            pass

        elif state == 5:
            self.updateTrayIconMenu()

        self.trayIconMenu.setUpdatesEnabled(True);

    def getClipboard(self):
        return toUnicode(self.systemClipboard.text())

    def putClipboard(self, clip):
        self.systemClipboard.setText(clip)

    def factory(self, clip):
        return lambda: self.copySelectionToClipboard(clip)

    def copyNewEntryToClipboard(self):
        if len(self.clipActions) > 0:
            self.clipActions[0].trigger()

    def copySelectionToClipboard(self, clip):
        curr = self.getClipboard()
        if clip != curr:
            self.putClipboard(clip)

    def showInfoMessage(self, message, title='OmniCopy'):
        self.trayIcon.showMessage(title, message, Q.QSystemTrayIcon.Information, 6000)

    def linkDb(self):
        self.api.tryAuth()
        ret = self.api.link()
        if not ret == True:
            self.showInfoMessage(_(
                'Please locate your browser window and enter your credentials to link Dropbox with OmniCopy. If you don\'t have an account yet, you can create one now (it\'s free!)'))
            self.openURL(ret)
        else:
            self.updateThread.setStatus(1)

    def openURL(self, url):
        Q.QDesktopServices.openUrl(QUrl(url, QUrl.TolerantMode))

    def link(self):
        self.api.tryAuth()
        ret = self.api.link()

    def unlink(self):
        self.api.unlink()
        pass

    def clear(self):
        self.putClipboard("")

        for act in self.clipActions:
            self.trayIconMenu.removeAction(act)
            del act

        del self.clipActions[:]
        self.trayIconMenu.update()

        self.api.clear()

    def toggleAutostart(self):
        if autorun.exists(version.APP_NAME):
            autorun.remove(version.APP_NAME)
        else:
            autorun.add(version.APP_NAME, "omnicopy")
        self.setAutostart()

    def setAutostart(self):
        self.optionAutostart.setChecked(autorun.exists(version.APP_NAME))

    def aboutQt(self):
        Q.QMessageBox.aboutQt(None, _('OmniCopy is using Qt licensed under GNU LPGL version 2.1'))

    def onTrayIconActivated(self, reason):
        pos = Q.QCursor.pos()
        self.lastMousePos = pos

        if reason == Q.QSystemTrayIcon.Context or reason == Q.QSystemTrayIcon.Information:
            return

        self.trayIconMenu.activateWindow()
        self.trayIconMenu.popup(pos)

    def autoUpdate(self):
        self.serverVersion = self.checkUpdate()
        if self.newVersion == "":
            print "Could not check for new Version"
        elif self.newVersion == version.APP_VERSION:
            print "Current Version."
        else:
            print "Newer Version avaiable: %s" % self.newVersion
            self.aboutUpdate()

    @pyqtSlot()
    def openWebsite(self):
        self.openURL('http://www.omnicopy.de')

    @pyqtSlot()
    def checkUpdate(self):
        self.newVersion = ""
        self.labelNewVersion.setText(self.newVersion)
        text = ""
        try:
            httpServ = httplib.HTTPConnection("www.omnicopy.de", 80, timeout=2)
            httpServ.connect()
            version_file = "/version_%s.txt" % (platform.system().lower())
            httpServ.request('GET', version_file)
            response = httpServ.getresponse()
            if response.status == httplib.OK:
                self.newVersion = response.read().strip()
                self.labelNewVersion.setText(_('Current Version is %s') % self.newVersion)
                if self.newVersion != version.APP_VERSION:
                    self.about_homepage.setFocus()
                else:
                    self.about_close.setFocus()
            else:
                self.labelNewVersion.setText = _('Error requesting version information')

            httpServ.close()
            return self.newVersion
        except Exception as e:
            self.labelNewVersion.setText(_('Could not contact update server.'))
            return ""

    def about(self):
        if self.about_dialog is None:
            self.about_dialog = Q.QDialog(self)
            self.about_dialog.setWindowTitle(_('About OmniCopy'))

            self.about_dialog.setWindowFlags(self.about_dialog.windowFlags() & ~Qt.WindowContextHelpButtonHint | Qt.WindowStaysOnTopHint)

            log("Loading ui: " + RES("about.ui"))
            ui = uic.loadUi(RES("about.ui"), self.about_dialog)

            icon = Q.QIcon(RES("omnicopy_win.ico"))
            self.about_dialog.setWindowIcon(icon)

            ui.image.setPixmap(Q.QPixmap(RES("omnicopy_128.png")))

            ui.version.setText(_('Version %s') % version.APP_VERSION)
            ui.bla.setText(_(u'OmniCopy is developed and\n© 2013 by Bernd Eckstein.\nPlease visit the OmniCopy '
                             u'website.'))

            ui.close.setText(_('Close'))
            ui.close.connect(ui.close, SIGNAL('clicked()'), self.about_dialog, SLOT('close()'))

            ui.homepage.setText(_('Visit www.omnicopy.de'))
            ui.homepage.connect(ui.homepage, SIGNAL('clicked()'), self, SLOT('openWebsite()'))

            ui.check.setText(_('Check for update'))
            ui.check.connect(ui.check, SIGNAL('clicked()'), self, SLOT('checkUpdate()'))



            self.labelNewVersion = ui.new_version
            self.labelNewVersion.setText("")

            self.about_check = ui.check
            self.about_homepage = ui.homepage
            self.about_close = ui.close

        self.about_check.setVisible(True)
        self.about_check.setFocus()

        self.about_dialog.show()
        self.about_dialog.raise_()

    def aboutUpdate(self):
        self.about()
        self.checkUpdate()
        self.about_check.setVisible(False)

    def lastWindowClosed(self):
        log("Last Window closed.")

    def createActions(self):
        self.linkAction = Q.QAction(_('Link to Dropbox'), self, triggered=self.linkDb)
        self.unlinkAction = Q.QAction(_('Unlink from Dropbox'), self, triggered=self.unlink)
        self.clearAction = Q.QAction(_('Clear Clipboard'), self, triggered=self.clear)

        self.optionNotify = Q.QAction(_('Enable notifications'), self)
        self.optionNotify.setCheckable(True)

        self.optionAutostart = Q.QAction(_('Enable autostart'), self, triggered=self.toggleAutostart)
        self.optionAutostart.setCheckable(True)
        self.setAutostart()

        self.status = Q.QAction(_('Initializing'), self);
        self.status.setEnabled(False)

    def createTrayIcon(self):
        # Wait up to 5 secs for the System Tray (autostart needs that!)
        count = 50
        while not Q.QSystemTrayIcon.isSystemTrayAvailable():
            if count == 50:
                print("Waiting for system tray ...")
            count -= 1
            sleep(0.1)

        if Q.QSystemTrayIcon.isSystemTrayAvailable():
            log("System Tray available.")
        else:
            print("No System Tray available!")
            self.app.exit(1)

        self.trayIconMenu = Q.QMenu(self);
        self.insertPoint = self.trayIconMenu.addSeparator()
        # self.trayIconMenu.addSeparator()
        self.trayIconMenu.addAction(self.clearAction)
        # Submenu Options
        self.trayIconSubMenu = Q.QMenu(_('Options'), self);
        self.trayIconMenu.addMenu(self.trayIconSubMenu)
        self.trayIconSubMenu.addAction(self.linkAction)
        self.trayIconSubMenu.addAction(self.unlinkAction)
        self.trayIconSubMenu.addSeparator()
        self.trayIconSubMenu.addAction(self.optionNotify)
        self.trayIconSubMenu.addAction(self.optionAutostart)
        self.trayIconSubMenu.addSeparator()
        # self.trayIconSubMenu.addAction(_('About Qt ...'),  self.aboutQt)
        self.trayIconSubMenu.addAction(_('About OmniCopy ...'), self.about)

        self.trayIconMenu.addSeparator()
        self.trayIconMenu.addAction(self.status)
        self.trayIconMenu.addSeparator()
        self.trayIconMenu.addAction(_('Quit OmniCopy'), self.quit)

        self.trayIconMenu.setSeparatorsCollapsible(False)

        self.icon_dark = Q.QIcon()
        self.icon_bright = Q.QIcon()
        if (platform.system() == 'Windows'):
            self.icon_dark.addPixmap(Q.QPixmap(RES("win_indicator_icon.png")), Q.QIcon.Normal, Q.QIcon.On)
            self.icon_bright.addPixmap(Q.QPixmap(RES("win_indicator_icon_blink.png")), Q.QIcon.Normal, Q.QIcon.On)
        else:
            self.icon_dark.addPixmap(Q.QPixmap(RES("indicator_icon.png")), Q.QIcon.Normal, Q.QIcon.On)
            self.icon_bright.addPixmap(Q.QPixmap(RES("indicator_icon_blink.png")), Q.QIcon.Normal, Q.QIcon.On)

        self.trayIcon = Q.QSystemTrayIcon(self)
        self.trayIcon.setIcon(self.icon_dark)
        self.trayIcon.setContextMenu(self.trayIconMenu)
        self.trayIcon.activated.connect(self.onTrayIconActivated)
        self.image128 = os.path.realpath(RES("omnicopy-128.png"))


class UpdateThread(QThread):
    updateData = pyqtSignal(int)

    def __init__(self, api, sigNewClipboard):
        QThread.__init__(self)
        self.keeprunning = True
        self.revision = -1
        self.api = api
        self.status = 3
        self.newClip = None
        sigNewClipboard.connect(self.newClipboard)

    def newClipboard(self, clip):
        self.newClip = clip

    def setStatus(self, stat):
        self.status = stat

    def setRunning(self, run):
        self.keeprunning = run

    def run(self):
        while (self.keeprunning):
            for i in range(1, 14):
                self.usleep(10 * 1000)
                if not self.keeprunning:
                    return
            """
            if self.status != self.api.currentStatus():
                # Update State on state change
                log("State change: %s"%self.api.currentStatus())
                emitlater = True
            else:
                emitlater = False
                log("Status still: %d" % self.status)

            self.status = self.api.currentStatus()
            """
            log("Tick %d" % self.status)
            try:
                """
                httpServ = httplib.HTTPConnection("api.dropbox.com", 80)
                httpServ.connect()
                httpServ.close()
                log("Internet available")
                """

                if self.status == 0:
                    pass
                    #  self.updateData.emit(0)

                if self.status == 0 or self.status == 1 or self.status == 2:
                    if self.status == 2:  # No internet ... try to connect ... slower
                        self.usleep(1000 * 1000)
                    log("check cloud")
                    revision = self.api.checkCloud()
                    log(revision)

                elif self.status == 4:  # In linking phase ... link on tryAuth()
                    log("tryAuth()")
                    if self.api.tryAuth():
                        log("true ... self.api.link()")
                        ret = self.api.link()
                        log(ret)
                    else:
                        log("try auth false")

                if self.api.currentStatus() == 1 and self.newClip != None:
                    # Add new Clipboard if ...
                    ret = self.api.addNewEntry(self.newClip)
                    if ret:
                        self.updateData.emit(0)

                # if emitlater:
                self.updateData.emit(self.api.currentStatus())
                self.status = self.api.currentStatus()

            except Exception as e:
                print e
                print '-' * 60
                traceback.print_exc(file=sys.stdout)
                print '-' * 60
                log("UpdateThread: %s\nTHIS SHOULD NOT HAPPEN" % e)
                self.usleep(1000 * 1000)


def main2():
    main(None);


def main(args):
    try:
        # Change to local directory (to find all resouces on windows machines)
        # os.chdir(sys.prefix)

        # program's main code here
        app = Q.QApplication(sys.argv)
        w = OmniCopy(app)
        app.setQuitOnLastWindowClosed(False)
        exitcode = app.exec_()
        sys.exit(exitcode)

    except Exception as e:
        print e
        return 1  # exit on error
    else:
        return 0  # exit without error


if __name__ == '__main__':
    sys.exit(main(sys.argv))

# Main


"""

Ubuntu: allow omnicooy in taskbar
check:
gsettings get com.canonical.Unity.Panel systray-whitelist
set:
gsettings set com.canonical.Unity.Panel systray-whitelist "['JavaEmbeddedFrame', 'Wine', 'scp-dbus-service', 'Update-notifier', 'python', 'omnicopy']"




Autostart Windows:
QSettings settings("HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Run", QSettings::NativeFormat);

Autostart Linux:
$HOME/.config/autostart
-- OmniCopy.desktop -------------------------
[Desktop Entry]

Name=OmniCopy
GenericName=Clipboard Synchronizer
Comment=Sync your clipboard across different devices and operating systems
Exec=OmniCopy
Terminal=false
Type=Application
Icon=omicopy
Categories=Network;FileTransfer;
StartupNotify=false
---------------------------------------------

Autostart OSX
--- applescript-file ------------------------
tell application "System Events"
    make login item at end with properties {path:"/Applications/Dictionary.app", hidden:false,
    kind:application, name:"OmniCopy"}
 end tell

---------------------------------------------
Delete login item
---------------------------------------------
tell application "System Events"
    if login item "OmniCopy" exists then delete login item "OmniCopy"
end tell
---------------------------------------------


"""
